@@ -3,7 +3,7 @@ package Net::Async::Redis::Cluster::Node;
 use strict;
 use warnings;
 
-our $VERSION = '3.011'; # VERSION
+our $VERSION = '3.012'; # VERSION
 
 use parent qw(IO::Async::Notifier);
 
@@ -18,14 +18,15 @@ use overload
 
 sub configure {
     my ($self, %args) = @_;
-    for my $k (qw(start end primary replicas)) {
+    for my $k (qw(start end primary replicas), @Net::Async::Redis::Cluster::CONFIG_KEYS) {
         $self->{$k} = delete $args{$k} if exists $args{$k};
     }
+    Scalar::Util::weaken($self->{cluster} = delete $args{cluster}) if exists $args{cluster};
     return $self->next::method(%args);
 }
 
 sub from_arrayref {
-    my ($class, $arrayref) = @_;
+    my ($class, $arrayref, %args) = @_;
     my ($start, $end, $primary, @replicas) = $arrayref->@*;
     die 'invalid start' if $start > $end or $start < 0 or $start >= 16384;
     die 'invalid end' if $end >= 16384;
@@ -36,6 +37,7 @@ sub from_arrayref {
         end      => $end,
         primary  => $primary,
         replicas => \@replicas,
+        %args
     )
 }
 
@@ -47,19 +49,27 @@ sub replica_list { shift->{replicas}->@* }
 sub replica_count { 0 + shift->{replicas}->@* }
 sub id { $_[0]->{primary}[2] // $_[0]->host_port }
 sub host_port { join ':', @{$_[0]->{primary}}[0, 1] }
+sub cluster { shift->{cluster} }
 
-async sub primary_connection {
+sub primary_connection {
     my ($self) = @_;
-    return $self->{primary_connection} //= do {
-        $self->add_child(
-            my $redis = Net::Async::Redis->new(
-                host => $self->primary->[0],
-                port => $self->primary->[1],
-            )
-        );
-        await $redis->connected;
-        $redis
-    };
+    return $self->{primary_connection} ||= $self->establish_primary_connection;
+}
+
+async sub establish_primary_connection {
+    my ($self) = @_;
+
+    $self->add_child(
+        my $redis = Net::Async::Redis->new(
+            $self->Net::Async::Redis::Cluster::node_config,
+            host => $self->primary->[0],
+            port => $self->primary->[1],
+        )
+    );
+    $self->{primary_connection} = $redis;
+    await $redis->connected;
+    await $self->cluster->node_connection_established($self, $redis);
+    return $redis;
 }
 
 1;
